/**
* wrt_conf_srv.c - Linksys WRT54G/GS configuration server
* Used for remotely configuring NVRAM settings 
*
* Copyright (C) 2004-2005 Pantelis Frangoudis, Elias Efstathiou
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdlib.h>
#include <netdb.h>
#include <unistd.h>
#include <string.h>
#include <bcmnvram.h>

#define ACTION_GET 10
#define ACTION_SET 20
#define ACTION_UNSET 30
#define ACTION_ACK 40
#define ACTION_ERR 50
#define ACTION_AUTH 60
#define ACTION_AUTH_OK 70
#define ACTION_AUTH_ERR 80

#ifndef CONF_SRV_PORT
#define CONF_SRV_PORT 5555
#endif

typedef struct key_value_pair_t {
	unsigned char *var;
	unsigned char *val;
} VPAIR;

typedef struct msg_t {
	unsigned short mtype;
	unsigned char *mtext;
} MSG;

int set_var(unsigned char *var, unsigned char *val) {
	int retval = 0;
	if (!var || !val) {
		return -1;
	}
	
	if(!nvram_get(var) || strcmp(nvram_get(var), val)) {
		retval = nvram_set(var, val);
	}
	else {
		return -1;
	}
	
	
	if (retval == 0) {
		return nvram_commit();
	}
	else {
		return -1;
	}
}

int unset_var(unsigned char *var) {
	int retval = 0;
	if (!var) {
		return -1;
	}
	
	if(nvram_get(var)) {
		retval = nvram_unset(var);
	}
	else {
		return -1;
	}

	if (retval == 0) {	
		return nvram_commit();
	}
	else {
		return -1;
	}
}


unsigned char *get_var(unsigned char *var) {
	unsigned char *nvram_var = NULL;
	unsigned char *retval = NULL;
	if (!var) {
		return NULL;
	}
	nvram_var = nvram_safe_get(var);
	if (strlen(nvram_var) == 0) {
		return NULL;
	}
	else {
		retval = (unsigned char*)malloc(strlen(nvram_var) + 1);
		memset(retval, 0, strlen(nvram_var) + 1);
		strncpy(retval, nvram_var, strlen(nvram_var));
	}
	return retval;
}

unsigned char *make_msg(unsigned short action, unsigned char *var, unsigned char *val) {
	unsigned char *text;
	int len = 0;
	int clen = 0;
	unsigned char strcl[10];
	memset(strcl, 0, 10);

	len = 3 + strlen("\r\nContent-length: ");
	switch (action) {
		case ACTION_ACK:
			clen += strlen("VAR: ") + strlen(var);
			sprintf(strcl, "%d", clen);
			len += strlen(strcl) + clen + 3; //add length of string repr of clen (+3: 2 for the \r\n and 1 xtra null byte - string end
			text = (unsigned char*)malloc(len);
			memset(text, 0, len);
			sprintf(text, "ACK\r\nContent-length: %d\r\nVAR: %s", clen, var);
			return text;

		case ACTION_ERR:
			clen += strlen("VAR: ") + strlen(var);
			sprintf(strcl, "%d", clen);
			len += strlen(strcl) + clen + 3; //add length of string repr of clen (+3: 2 for the \r\n and 1 xtra null byte - string end
			text = (unsigned char*)malloc(len);
			memset(text, 0, len);
			sprintf(text, "ERR\r\nContent-length: %d\r\nVAR: %s", clen, var);
			return text;

		case ACTION_AUTH_OK:
			len = 4;
			text = (unsigned char*)malloc(len);
			memset(text, 0, len);
			strcpy(text, "AOK");
			return text;

		case ACTION_AUTH_ERR:
			len = 4;
			text = (unsigned char*)malloc(len);
			memset(text, 0, len);
			strcpy(text, "AER");
			return text;

		default:
			return NULL;
	}
}

VPAIR *parse_msg(const unsigned char *text) {
	unsigned char text_copy[strlen(text)+1];
	memset(text_copy, 0, strlen(text) + 1);
	strcpy(text_copy, text);
	VPAIR *pair = NULL;
	int clen = 0;
	unsigned char *pos = NULL;
	unsigned char *crlfpos = NULL;
	unsigned char strclen[10];
	memset(strclen, 0, 10);

	if (
		!strstr(text, "GET\r\nContent-length: ") &&
		!strstr(text, "SET\r\nContent-length: ") &&
		!strstr(text, "UST\r\nContent-length: ") &&
		!strstr(text, "AUT\r\nContent-length: ")
		){
		return NULL;
	}
	pos = text_copy + 3 + strlen("\r\nContent-length: ");
	crlfpos = strstr(pos, "\r\n");
	if (!crlfpos) {
		return NULL;
	}
	strncpy(strclen, pos, crlfpos - pos);
	clen = atoi(strclen);
	pos = crlfpos + 2;

	if (strncasecmp(pos, "VAR: ", 5)) {
		return NULL;
	}
	pos += 5;

	pair = (VPAIR*)malloc(sizeof(VPAIR));
	memset(pair, 0, sizeof(VPAIR));

	if ( !(crlfpos = strstr(pos, "\r\n")) ) {
		//"get" or "ust"  message
		pair->var = (unsigned char*)malloc(clen - strlen("VAR: ") + 1);
		memset(pair->var, 0, clen - strlen("VAR: ") + 1);
		strncpy(pair->var, pos, clen - strlen("VAR: "));
		return pair;
	}

	//else: "SET" message
	pair->var = (unsigned char*)malloc(crlfpos - pos + 1); //+1: final '\0'
	memset(pair->var, 0, crlfpos - pos + 1);
	strncpy(pair->var, pos, crlfpos - pos);
	pos = crlfpos + 2;
	if (strncasecmp(pos, "VAL: ", 5)) {
		//malformed message
		if (pair->var) free(pair->var);
		if (pair) free(pair);
		return NULL;
	}
	pos += 5;
	pair->val = (unsigned char*)malloc(clen - strlen(pair->var) - 11);
	memset(pair->val, 0, clen - strlen(pair->var) - 11);
	strncpy(pair->val, pos, clen - strlen(pair->var) - 12);
	return pair;
}

MSG *read_message(int sockfd) {
	//first read the first 4 bytes (message type) - check if its a conf msg, else exit
	unsigned char *text_message = NULL;
	MSG *retval = NULL;

	unsigned char *header = NULL; //content length header (this array is read from the socket)

	//message type string(read from the socket)
	unsigned char strtype[4];
	memset(strtype, 0, 4);

	unsigned char strcl[10]; //content length of message
	int clen = 0;
	int bytesread = 0;
	int totlen = 0; //total len of the text message
	int mtype = 0;
	int i = 0;

	//read message type (the first 3 bytes)
	bytesread = read(sockfd, strtype, 3);
	if (bytesread != -1) {
		//save message type
		if (!strncasecmp(strtype, "SET", 3)) {
			mtype = ACTION_SET;
		}
		else
		if (!strncasecmp(strtype, "GET", 3)) {
			mtype = ACTION_GET;
		}
		else
		if (!strncasecmp(strtype, "UST", 3)) {
			mtype = ACTION_UNSET;
		}
		else
		if (!strncasecmp(strtype, "AUT", 3)) {
			mtype = ACTION_AUTH;
		}
		else {
			fprintf(stderr, "Unknown message type\n");
			return NULL;
		}

		header = (unsigned char*)malloc(strlen("\r\nContent-length: ") + 1);
		memset(header, 0, strlen("\r\nContent-length: ") + 1);
		bytesread = read(sockfd, header, strlen("\r\nContent-length: "));
		if (bytesread != -1) {
			//first check header
			if (strcasecmp(header, "\r\nContent-length: ")) {
				//malformed header
				if (header) free(header);
				return NULL;
			}
			//read content length
			//actually, we read byte after byte until we find a crlf
			memset(strcl, 0, 10);
			i = 0;
			while ( !strstr(strcl, "\r\n") && (i < 10) ) {
				if (read(sockfd, &strcl[i], 1) == -1) {
					//error reading character
					if (header) free(header);
					return NULL;
				}
				i++;
			}
			if (!sscanf(strcl, "%d\r\n", &clen)) {
				//malformed clen
				if (header) free(header);
				return NULL;
			}

			//finally, read the remaining text message into "text_message" (which will eventually be returned)
			totlen = 3 + strlen(header) + strlen(strcl) + clen + 1; //total length of message ("3": mtype, "+1": trailing \0)
			text_message = (unsigned char*)malloc(totlen);
			memset(text_message, 0, totlen);

			//copy the fields that are already read
			sprintf(text_message, "%s%s%s", strtype, header, strcl);
			//"append" the remaining bytes to text_message (start writing from position text_message+sizeof(already read))
			bytesread = read(sockfd, text_message + strlen(text_message), clen);
			if (bytesread != clen) {
				//error reading data from socket: bytesread must be == with the calculated text message length
				if(text_message) free(text_message);
				if (header) free(header);
				return NULL;
			}
		}
		else {
			if (header) free(header);
			return NULL;
		}
	}
	else {
		return NULL;
	}
	if(header) free(header);
	retval = (MSG*)malloc(sizeof(MSG));
	memset (retval, 0, sizeof(MSG));
	retval->mtype = mtype;
	retval->mtext = (unsigned char*)malloc(strlen(text_message) + 1);
	memset(retval->mtext, 0, strlen(text_message) + 1);
	strcpy(retval->mtext, text_message);
	if (text_message) free(text_message);

	return retval;
}

int auth_user(int sockfd) {
	MSG *msg = NULL;
	VPAIR *passwd = NULL;
	unsigned char *nvram_passwd = NULL;
	unsigned char *resp = NULL;
	msg = read_message(sockfd);
	if (!msg) {
		return 0;
	}
	if (!msg->mtext) {
		if (msg) {
			free(msg);
		}
		return 0;
	}
	if (msg->mtype != ACTION_AUTH) {
		if (!msg->mtext) {
			if (msg) {
				free(msg);
			}
		}
		return 0;
	}

	passwd = parse_msg(msg->mtext);
	if (!passwd) {
		return 0;
	}

	
	//read HTTP password from NVRAM
	//this is the passwd used for authentication
	nvram_passwd = get_var("http_passwd");
	if (nvram_passwd) {
		if (!strcmp(nvram_passwd, passwd->var)) {
			//passwords match
			resp = make_msg(ACTION_AUTH_OK, NULL, NULL);
		}
		else {
			//passwords do not match
			resp = make_msg(ACTION_AUTH_ERR, NULL, NULL);
		}
	}
	else {
		//error retrieving passwd from nvram
		resp = make_msg(ACTION_AUTH_ERR, NULL, NULL);
	}
	
	//free passwd struct
	if (passwd) {
		if (passwd->var) {
			free(passwd->var);
		}
		free(passwd);
	}

	//free the value returned from get_var
	if (nvram_passwd) {
		free(nvram_passwd);
		nvram_passwd = NULL;
	}
	
	//send resp to client
	if (!write(sockfd, resp, strlen(resp))) {
		if (resp) {
			free(resp);
		}
		return 0;
	}

	return 1;
}

int main(int argc, char **argv) {
	struct sockaddr_in server_addr;
	int sockfd, sockfd2;
	struct sockaddr_in client_addr;
	int alen = sizeof(client_addr);

	memset(&server_addr, 0, sizeof(server_addr));
	server_addr.sin_family = AF_INET;
	server_addr.sin_addr.s_addr = INADDR_ANY;
	server_addr.sin_port = htons(CONF_SRV_PORT);

	sockfd = socket (PF_INET, SOCK_STREAM, 0);
	if (sockfd < 0) {
		fprintf(stderr, "socket creation failed\n");
		return 0;
	}

	//Bind a local address to the socket
	if (bind(sockfd, (struct sockaddr *)&server_addr, sizeof (server_addr)) < 0) {
		fprintf(stderr,"bind failed\n");
		return 0;
	}

	if (listen(sockfd, 5) < 0) {
		fprintf(stderr,"listen failed\n");
		return 0;
	}

	VPAIR *pair;
	MSG *msg;
	unsigned char *resp = NULL;
	unsigned char *to_be_got = NULL;

	while (1) {
		if (  (sockfd2=accept(sockfd, (struct sockaddr *)&client_addr, &alen)) < 0) {
			fprintf(stderr, "accept failed\n");
			exit (1);
		}

		//authenticate user
		if (!auth_user(sockfd2)) {
			close(sockfd2);
			fprintf(stderr, "User not authenticated\n");
			continue;
		}
		msg = read_message(sockfd2);
		if (!msg) {
			fprintf(stderr, "null message\n");
		}
		else {
			pair = parse_msg(msg->mtext);
			switch (msg->mtype) {
				case ACTION_SET:
					if (set_var(pair->var, pair->val)) {
						resp = make_msg(ACTION_ERR, pair->var, NULL);
					}
					else {
						resp = make_msg(ACTION_ACK, pair->var, NULL);
					}
					break;
				case ACTION_GET:
					to_be_got = get_var(pair->var);
					if (to_be_got) {
						resp = make_msg(ACTION_ACK, to_be_got, NULL);
					}
					else {
						resp = make_msg(ACTION_ERR, pair->var, NULL);
					}
					if (to_be_got) {
						free(to_be_got);
						to_be_got = NULL;
					}
//					printf("++%d\n", get_var(pair->var));
					break;
				case ACTION_UNSET:
					if (unset_var(pair->var)) {
						resp = make_msg(ACTION_ERR, pair->var, NULL);
					}
					else {
						resp = make_msg(ACTION_ACK, pair->var, NULL);
					}
					break;
			}
			if (!write(sockfd2, resp, strlen(resp))) {
				fprintf(stderr, "Error writing to socket\n");
				if (resp) free(resp);
				if (pair) {
					if (pair->var) free(pair->var);
					if (pair->val) free(pair->val);
					free(pair);
				}
				return 1;
			}
			if (resp) {
				free(resp);
				resp = NULL;
			}
			if (pair) {
				if (pair->var) free(pair->var);
				if (pair->val) free(pair->val);
				free(pair);
			}
		}
		close(sockfd2);
		if (msg) {
			if (msg->mtext) {
				free(msg->mtext);
				msg->mtext = NULL;
			}
			free(msg);
			msg = NULL;
		}
	}
	close(sockfd);
	return 0;
}
